package Game
{
	import starling.core.Starling;
	import starling.display.Image;
	import starling.display.Sprite;
	import starling.events.EnterFrameEvent;
	import starling.events.Event;
	import starling.text.BitmapFont;
	import starling.text.TextField;
	
	public class GameScene extends Sprite
	{
		public var ui:GameUI;
		public var fx:GameFX;
		public var bg:Background;
		public var mainLayer:Sprite;
		public var shipLayer:Sprite;
		public var trailLayer:Sprite;
		public var particleLayer:Sprite;
		public var hero:Hero;
		
		public var ships:Array;
		public var shipPool:Array;
		public var trails:Array;
		public var trailPool:Array;
		public var projectiles:Array;
		public var projectilePool:Array;
		public var particles:Array;
		public var particlePool:Array;
		public var powerups:Array;
		public var powerupPool:Array;
		
		public var gameTimer:Number;
		public var survived:Number;
		public var kills:int;
		public var bestSurvived:Number;
		public var bestKills:int;
		public var maxTime:Number;
		public var waves:int;
		
		public var logo:Image;
		public var credits:TextField;
		public var results:TextField;
		public var gameOver:Boolean;
		
		public function GameScene()
		{
			bg = new Background(this);
			addChild(bg);
			
			shipLayer = new Sprite();
			trailLayer = new Sprite();
			particleLayer = new Sprite();
			mainLayer = new Sprite();
			mainLayer.addChild(trailLayer);
			mainLayer.addChild(particleLayer);
			mainLayer.addChild(shipLayer);
			addChild(mainLayer);
			
			ships = [];
			shipPool = [];
			trails = [];
			trailPool = [];
			projectiles = [];
			projectilePool = [];
			particles = [];
			particlePool = [];
			powerups = [];
			powerupPool = [];
			
			hero = new Hero(this);
			ui = new GameUI(this);
			fx = new GameFX(this);
			
			logo = new Image(Root.assets.getTexture("logo"));
			logo.pivotX = logo.width*0.5;
			logo.pivotY = logo.height*0.5;
			logo.color = 0x00aaff;
			logo.x = 400;
			logo.y = 300;
			logo.blendMode = "add";
			addChild(logo);
			
			credits = new TextField(800, 20, "Created in 48 hours for Ludum Dare 27 (August 23-25, 2013) by Nico Tuason".toUpperCase(), "Arial", 11, 0x006699, true);
			credits.hAlign = "center";
			credits.vAlign = "top";
			credits.x = 0;
			credits.y = 580;
			credits.blendMode = "add";
			addChild(credits);
			
			results = new TextField(800, 600, "GAME OVER", "Downlink", BitmapFont.NATIVE_SIZE, 0x00aaff);
			results.hAlign = "center";
			results.vAlign = "center";
			results.blendMode = "add";
			addChild(results);
			
			bestSurvived = 0;
			bestKills = 0;
		}
		
		public function init():void {
			gameOver = true;
			waves = 0;
			
			mainLayer.scaleX = mainLayer.scaleY = 0.5;
			bg.init();
			hero.init();
			ui.init();
			fx.init();
			results.alpha = 0;
			addEventListener(Event.ENTER_FRAME, update);
		}
		
		public function reset():void {
			ui.reset();
			Starling.juggler.tween(results, 0.5, {alpha:0});
			Starling.juggler.tween(logo, 0.5, {alpha:1, delay:0.5});
			Starling.juggler.tween(credits, 0.5, {alpha:1, delay:0.5});
		}
		
		public function start():void {
			gameTimer = 60;
			survived = 0;
			kills = 0;
			waves = 0;
			ui.start();
			Starling.juggler.delayCall(function():void{gameOver = false;addWave();}, 0.5);
			Starling.juggler.tween(logo, 0.5, {alpha:0});
			Starling.juggler.tween(credits, 0.5, {alpha:0});
		}
		
		public function deInit():void {
			hero.deInit();
			ui.deInit();
			fx.deInit();
			removeEventListener(Event.ENTER_FRAME, update);
		}
		
		public function showGameOver():void {
			clearObjects();
			if(survived > bestSurvived) bestSurvived = survived;
			if(kills > bestKills) bestKills = kills;
			
			results.text = "GAME OVER\n\n\nYou survived for: " + getTimeStr(survived) + "\n\nRecord: " + getTimeStr(bestSurvived) + "\n\n\nKill count: " + kills + "\n\nRecord: " + bestKills;
			
			Starling.juggler.tween(results, 2, {alpha:1});
		}
		
		public function getTimeStr(time:Number):String {
			var secs:String = int(time).toString();
			if(secs.length == 1) secs = "0" + secs;
			var mils:String = int((time*100)%100).toString();
			if(mils.length == 1) mils = "0" + mils;
			return(secs + ":" + mils);
		}
		
		public function clearObjects():void {
			for(var i:int = 0; i < ships.length; i++) ships[i].active = false;
			for(i = 0; i < projectiles.length; i++) projectiles[i].active = false;
			for(i = 0; i < particles.length; i++) particles[i].active = false;
			for(i = 0; i < powerups.length; i++) powerups[i].active = false;
		}
		
		public var dt:Number;
		public var dtMult:Number = 1;
		public function update(e:EnterFrameEvent):void {
			dt = e.passedTime;
			dt *= dtMult;
			
			if(!gameOver) {
				gameTimer -= dt;
				survived += dt;
				if(gameTimer <= 0) {
					showGameOver();
					gameOver = true;
				}
			}
			
			hero.update(dt);
			
			updateCamera(dt);
			updateShips(dt);
			updateProjectiles(dt);
			updateParticles(dt);
			updateTrails(dt);
			updatePowerups(dt);
			
			fx.update(dt);
			ui.update(dt);
			GS.update(dt);
		}
		
		public function updateCamera(dt:Number):void {
			var dx:Number = (-hero.x+400*2) - mainLayer.x*2;
			var dy:Number = (-hero.y+300*2) - mainLayer.y*2;
			var dist:Number = Math.sqrt(dx*dx + dy*dy);
			var angle:Number = Math.atan2(dy, dx);
			var scrollX:Number = Math.cos(angle)*dist*0.05;
			var scrollY:Number = Math.sin(angle)*dist*0.05;
			mainLayer.x += scrollX;
			mainLayer.y += scrollY;
			ui.hud.x = -scrollX*2;
			ui.hud.y = -scrollY*2;
			bg.update(dt, scrollX, scrollY);
		}
		
		public function updateShips(dt:Number):void {
			var l:int = ships.length;
			for(var i:int = 0; i < l; i++) {
				var ship:Ship = ships[i];
				if(!ship.active) {
					ship.deInit();
					shipPool.push(ship);
					ships[i] = ships[l-1];
					ships.pop();
					l--; i--;
				} else ship.update(dt);
			}
		}
		
		public function updateProjectiles(dt:Number):void {
			var l:int = projectiles.length;
			for(var i:int = 0; i < l; i++) {
				var projectile:Projectile = projectiles[i];
				if(!projectile.active) {
					projectile.deInit();
					projectilePool.push(projectile);
					projectiles[i] = projectiles[l-1];
					projectiles.pop();
					l--; i--;
				} else projectile.update(dt);
			}
		}
		
		public function updateParticles(dt:Number):void {
			var l:int = particles.length;
			for(var i:int = 0; i < l; i++) {
				var particle:ParticleFX = particles[i];
				if(!particle.active) {
					particle.deInit();
					particlePool.push(particle);
					particles[i] = particles[l-1];
					particles.pop();
					l--; i--;
				} else particle.update(dt);
			}
		}
		
		public function updateTrails(dt:Number):void {
			var l:int = trails.length;
			for(var i:int = 0; i < l; i++) {
				var trail:TrailFX = trails[i];
				if(!trail.active) {
					trail.deInit();
					trailPool.push(trail);
					trails[i] = trails[l-1];
					trails.pop();
					l--; i--;
				} else trail.update(dt);
			}
		}
		
		public function updatePowerups(dt:Number):void {
			var l:int = powerups.length;
			for(var i:int = 0; i < l; i++) {
				var powerup:PowerUp = powerups[i];
				if(!powerup.active) {
					powerup.deInit();
					powerupPool.push(powerup);
					powerups[i] = powerups[l-1];
					powerups.pop();
					l--; i--;
				} else powerup.update(dt);
			}
		}
		
		public function addWave():void {
			var l:int = waves+3;
			for(var i:int = 0; i < l; i++) {
				var angle:Number = Math.random()*Math.PI;
				if(Math.random() < 0.5) angle *= -1;
				var dx:Number = hero.x + Math.cos(angle)*900;
				var dy:Number = hero.y + Math.sin(angle)*900;
				addShip(dx, dy);
				ships[ships.length-1].hasPowerUp = true;
			}
			waves++;
		}
		
		public function addShip(x:Number, y:Number):void {
			var ship:Ship;
			if(shipPool.length > 0) ship = shipPool.pop();
			else ship = new Ship(this);
			ship.init(x, y);
			ships.push(ship);
		}
		
		public function addPowerup(x:Number, y:Number, vector:NumPoint):void {
			var powerup:PowerUp;
			if(powerupPool.length > 0) powerup = powerupPool.pop();
			else powerup = new PowerUp(this);
			powerup.init(x, y, vector);
			powerups.push(powerup);
		}
		
		public function addProjectile(x:Number, y:Number, vector:NumPoint, rotation:Number, team:int, color:uint):void {
			var projectile:Projectile;
			if(projectilePool.length > 0) projectile = projectilePool.pop();
			else projectile = new Projectile(this);
			projectile.init(x, y, vector, rotation, team, color);
			projectiles.push(projectile);
			GS.playLaser();
		}
		
		public function addParticle():void {
			
		}
		
		public function addHitParticles(x:Number, y:Number, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initHit(x, y, color);
			particles.push(particle);
		}
		
		public function addShockParticle(x:Number, y:Number, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initShock(x, y, color);
			particles.push(particle);
		}
		
		public function addSmoke(x:Number, y:Number, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initSmoke(x, y, color);
			particles.push(particle);
		}
		
		public function addShard(x:Number, y:Number, vector:NumPoint, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initShard(x, y, vector, color);
			particles.push(particle);
		}
		
		public function addFlare(x:Number, y:Number, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initFlare(x, y, color);
			particles.push(particle);
		}
		
		public function addFlare2(x:Number, y:Number, color:uint):void {
			var particle:ParticleFX;
			if(particlePool.length > 0) particle = particlePool.pop();
			else particle = new ParticleFX(this);
			particle.initFlare2(x, y, color);
			particles.push(particle);
		}
		
		public function addTrail(x:Number, y:Number, rotation:Number, width:Number, color:uint=0xffffff):TrailFX {
			var trail:TrailFX;
			if(trailPool.length > 0) trail = trailPool.pop();
			else trail = new TrailFX();
			trail.initTrail(x, y, width, rotation, color);
			trails.push(trail);
			return trail;
		}
		
		public function addTime():void {
			GS.playPowerup();
			gameTimer += 20;
			if(gameTimer > 60) gameTimer = 60;
			bg.highlight();
		}
		
		public function minusTime():void {
			gameTimer -= 2;
			if(gameTimer < 0) gameTimer = 0;
			bg.showHit();
		}
		
		public function checkNextWave():void {
			if(gameOver) return;
			if(ships.length <= 0) addWave();
		}
	}
}